@@ -5,6 +5,7 @@ from deepmerge import always_merger
 from delphai_utils.logging import default_config, logging
 from abc import abstractmethod, abstractproperty
 import faust
+import socket
 from dataclasses import dataclass, field
 from typing import Any, Dict, List, Optional, TypeVar, Type
 from delphai_utils.config import get_config
@@ -20,8 +21,37 @@ import coloredlogs
 from aioprometheus import Service, Counter, Summary
 from timeit import default_timer
 
+
 logging.getLogger('aiohttp').setLevel(logging.ERROR)
 
+def is_port_free(host, port):
+  """
+  determine whether `host` has the `port` free
+
+  From: https://www.thepythoncode.com/article/make-port-scanner-python
+  """
+  s = socket.socket()
+  try:
+    s.connect((host, port))
+  except Exception:
+    return True
+  else:
+    return False
+
+
+class NoPortFoundError(Exception):
+  ...
+
+
+def find_free_port(start: int, host='127.0.0.1', num_tries=4) -> int:
+  for port in range(start, start + num_tries):
+    if is_port_free(host, port):
+      return port
+    else:
+      logging.info(f'Port {port} already in use.')
+  message = f'No free port found in range [{start}, {start + num_tries - 1}]'
+  logging.error(message)
+  raise NoPortFoundError(message)
 
 @dataclass
 class Step:
@@ -198,8 +228,9 @@ class FaustApp():
       tasks = []
       for agent in agents:
         tasks.append(self.attach(agent))
-      asyncio.get_event_loop().create_task(self.metrics.start(port=9191))
-      self.logger.info('started metrics server on port 9191')
+      metrics_port = find_free_port(9191)
+      asyncio.get_event_loop().create_task(self.metrics.start(port=metrics_port))
+      self.logger.info(f'started metrics server on port {metrics_port}')
       self.worker.loop.run_until_complete(asyncio.gather(*tasks))
       self.worker.execute_from_commandline()
     except KeyboardInterrupt:
